#include <stdio.h>
#include <string.h>
#include <gtk/gtk.h>
#include <ecl/ecl.h>

/* The text view widget and its buffer / end-of-buffer mark */
GtkWidget *view;
GtkTextBuffer *buffer;
GtkTextMark *eob;

static void destroy(GtkWidget *widget, gpointer data)
{
  fprintf(stderr, "Shutting down...\n");
  cl_shutdown();
  gtk_main_quit();
}

static void evaluate(GtkEntry *entry, gpointer data)
{
  char const *text = gtk_entry_get_text(entry);
  cl_object result;

  /* Print the command */
  gtk_text_buffer_insert_at_cursor(buffer, "\n;;; ", 5);
  gtk_text_buffer_insert_at_cursor(buffer, text, strlen(text));
  gtk_text_buffer_insert_at_cursor(buffer, "\n", 1);

  /* Evaluate and print the result */
  result = cl_safe_eval(c_string_to_object(text), Cnil, Cnil);
  cl_princ(1, result);
  cl_terpri(0);

  /* Scroll the window */
  gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(view), eob, 0, FALSE, 0, 0);

  /* Clear the input widget */
  gtk_entry_set_text(entry, "");
}

static cl_object cl_write_text_view_char(int narg, cl_object ch)
{
  char c;
  check_arg(1);

  c = ecl_to_char(ch);
  gtk_text_buffer_insert_at_cursor(buffer, &c, 1);
  gtk_text_view_scroll_to_mark(GTK_TEXT_VIEW(view), eob, 0, FALSE, 0, 0);

  NVALUES = 1;
  return Cnil;
}

inline
cl_object symbol(char *str)
{
  return cl_intern(1, make_simple_base_string(str));
}

int main(int argc, char *argv[])
{
  PangoFontDescription *font;
  GtkTextIter end;
  GtkWidget *window, *box, *scrolled, *entry;

  fprintf(stderr, "Initializing...\n");

  /* Initialize Gtk */
  gtk_init(&argc, &argv);

  /* Create the widgets */
  window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_title(GTK_WINDOW(window), "ECL / Gtk Test");
  gtk_window_set_default_size(GTK_WINDOW(window), 500, 300);
  box = gtk_vbox_new(FALSE, 3);
  scrolled = gtk_scrolled_window_new(NULL, NULL);
  gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(scrolled),
				 GTK_POLICY_NEVER, GTK_POLICY_AUTOMATIC);
  buffer = gtk_text_buffer_new(NULL);
  gtk_text_buffer_get_end_iter(GTK_TEXT_BUFFER(buffer), &end);
  eob = gtk_text_buffer_create_mark(GTK_TEXT_BUFFER(buffer),
				    "EOB", &end, FALSE);
  view = gtk_text_view_new_with_buffer(buffer);
  gtk_text_view_set_editable(GTK_TEXT_VIEW(view), FALSE);
  font = pango_font_description_from_string("monospace 10");
  gtk_widget_modify_font(view, font);
  pango_font_description_free(font);
  entry = gtk_entry_new();

  /* Add and show the widgets */
  gtk_box_pack_start(GTK_BOX(box), scrolled, TRUE, TRUE, 0);
  gtk_widget_show(view);
  gtk_container_add(GTK_CONTAINER(scrolled), view);
  gtk_widget_show(scrolled);
  gtk_box_pack_start(GTK_BOX(box), entry, FALSE, TRUE, 0);
  gtk_widget_show(entry);
  gtk_container_add(GTK_CONTAINER(window), box);
  gtk_widget_show(box);
  gtk_widget_show(window);
  gtk_window_set_focus(GTK_WINDOW(window), entry);

  /* Set up the signals */
  g_signal_connect(G_OBJECT(window), "destroy", G_CALLBACK(destroy), NULL);
  g_signal_connect_swapped(G_OBJECT (window), "delete_event",
			   G_CALLBACK(gtk_widget_destroy),
			   G_OBJECT(window));
  g_signal_connect(G_OBJECT(entry), "activate", G_CALLBACK(evaluate), NULL);

  /* Initialize Lisp */
  cl_boot(argc, argv);
  cl_def_c_function_va(symbol("WRITE-TEXT-VIEW-CHAR"), /* uppercase! */
		       (void *)cl_write_text_view_char);
  cl_load(1, make_simple_base_string("init.lisp"));

  /* Start the GUI */
  gtk_main();

  return 0;
}
